/* The r250 uniform random number algorithm Kirkpatrick, S., and
   E. Stoll, 1981; "A Very Fast Shift-Register Sequence Random Number
   Generator", Journal of Computational Physics, V.40 also: see
   W.L. Maier, DDJ May 1991

   Modified by A. Kuronen, April 2003

*/

#include <math.h>
#include <limits.h>
#include "r250.h"

/* defines to allow for 16 or 32 bit integers */
#define BITS 31

#if WORD_BIT == 32
#ifndef BITS
#define BITS	32
#endif
#else
#ifndef BITS
#define BITS    16
#endif
#endif

#if BITS == 31
#define MSB          0x40000000
#define ALL_BITS     0x7fffffff
#define HALF_RANGE   0x20000000
#define STEP         7
#endif

#if BITS == 32
#define MSB          0x80000000
#define ALL_BITS     0xffffffff
#define HALF_RANGE   0x40000000
#define STEP         7
#endif

#if BITS == 16
#define MSB         0x8000
#define ALL_BITS    0xffff
#define HALF_RANGE  0x4000
#define STEP        11
#endif

/* Global variables */
static unsigned int r250_buffer[ 250 ];
static int r250_index;
static int r250_quotient = INT_MAX / 16807;
static int r250_remainder = INT_MAX % 16807;
static int r250_seed_val = 1;


void r250_init(int sd)
{
  int j, k;
  unsigned int mask, msb;
  
  r250_set_seed( sd );
  
  r250_index = 0;
  for (j = 0; j < 250; j++)      /* fill r250 buffer with BITS-1 bit values */
    r250_buffer[j] = r250_lcg();
  
  for (j = 0; j < 250; j++)	 /* set some MSBs to 1 */
    if ( r250_lcg() > HALF_RANGE )
      r250_buffer[j] |= MSB;
  
  msb = MSB;	                 /* turn on diagonal bit */
  mask = ALL_BITS;	         /* turn off the leftmost bits */
  
  for (j = 0; j < BITS; j++) {
    k = STEP * j + 3;	         /* select a word to operate on */
    r250_buffer[k] &= mask;      /* turn off bits left of the diagonal */
    r250_buffer[k] |= msb;	 /* turn on the diagonal bit */
    mask >>= 1;
    msb  >>= 1;
  }
  
}

unsigned int r250()		 /* returns a random unsigned integer */
{
  register int	j;
  register unsigned int new_rand;
  
  if ( r250_index >= 147 )
    j = r250_index - 147;	 /* wrap pointer around */
  else
    j = r250_index + 103;
  
  new_rand = r250_buffer[ r250_index ] ^ r250_buffer[ j ];
  r250_buffer[ r250_index ] = new_rand;
  
  if ( r250_index >= 249 )	 /* increment pointer for next time */
    r250_index = 0;
  else
    r250_index++;
  
  return new_rand;
  
}


double dr250()		         /* returns a random double in range 0..1 */
{
  register int	j;
  register unsigned int new_rand;
  
  if ( r250_index >= 147 )
    j = r250_index - 147;	 /* wrap pointer around */
  else
    j = r250_index + 103;
  
  new_rand = r250_buffer[ r250_index ] ^ r250_buffer[ j ];
  r250_buffer[ r250_index ] = new_rand;
  
  if ( r250_index >= 249 )	 /* increment pointer for next time */
    r250_index = 0;
  else
    r250_index++;
  
  return (double)new_rand / ALL_BITS;
  
}


/* Linear Congruential Method, the "minimal standard generator"
   Park & Miller, 1988, Comm of the ACM, 31(10), pp. 1192-1201
*/

int r250_set_seed(int sd) { return r250_seed_val = sd; }
int r250_get_seed() { return r250_seed_val; }

unsigned int r250_lcg()
{
  if ( r250_seed_val <= r250_quotient ) {
    r250_seed_val = (r250_seed_val * 16807) % INT_MAX;
  } else {
    int high_part = r250_seed_val / r250_quotient;
    int low_part  = r250_seed_val % r250_quotient;
    int test = 16807 * low_part - r250_remainder * high_part;
    if ( test > 0 )
      r250_seed_val = test;
    else
      r250_seed_val = test + INT_MAX;
  }
  
  return r250_seed_val;
}
